==> ocrad-0.9.tar.bz2: ocrad-0.9/character_r12.cc <==
/*  GNU Ocrad - Optical Character Recognition program
    Copyright (C) 2003, 2004 Antonio Diaz Diaz.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <cstdio>
#include <vector>
#include "common.h"
#include "rectangle.h"
#include "ucs.h"
#include "block.h"
#include "blockmap.h"
#include "character.h"
#include "profile.h"
#include "feats.h"


// Recognizes 2 block characters.
// ij!%:;=?|
void Character::recognize12( const Charset & charset, int charbox_vcenter ) throw()
  {
  const Block & b1 = block( 0 );		// upper block
  const Block & b2 = block( 1 );		// lower block
  int a1 = b1.area();
  int a2 = b2.area();
  Features f1( b1 );
  Features f2( b2 );

  if( Ocrad::similar( a1, a2, 10 ) )
    {
    if( !b1.blocks() && !b2.blocks() &&
        2 * a1 > b1.size() && 2 * a2 > b2.size() )
      {
      if( width() > height() || Ocrad::similar( width(), height(), 40 ) )
        { add_guess( '=', 0 ); return; }
      if( Ocrad::similar( b1.width(), b1.height(), 20, 2 ) &&
          Ocrad::similar( b2.width(), b2.height(), 20, 2 ) )
        add_guess( ':', 0 );
      return;
      }
    return;
    }
  if( Ocrad::similar( a1, a2, 60 ) )
    {
    if( f1.test_solid( charbox_vcenter ) == '.' )
      {
      if( f2.test_solid( charbox_vcenter ) == '.' )
        { add_guess( ':', 0 ); return; }
      if( b2.height() > b1.height() ) { add_guess( ';', 0 ); return; }
      }
    if( charset.enabled( Charset::iso_8859_15 ) ||
        charset.enabled( Charset::iso_8859_9 ) )
      {
      int code = f2.test_solid( charbox_vcenter );
      if( code == '-' || code == '_' )
        { add_guess( UCS::PLUSMIN, 0 ); return; }
      }
    if( b1.includes_hcenter( b2 ) && b2.includes_hcenter( b1 ) )
      { if( b1.blocks() && b2.blocks() ) { add_guess( 'g', 0 ); return; } }
    if( b1.hcenter() < b2.hcenter() )
      {
      // Looks for merged 'fi'
      if( b2.height() > b2.width() && b1.hcenter() < b2.left() &&
          b1.includes_hcenter( b2 ) && 4 * b1.height() > 5 * b2.height() &&
          Ocrad::similar( b1.bottom()-b1.top(), b2.bottom()-b1.top(), 10 ) )
        {
        Character c2( b2 ); c2.recognize1( charset, vcenter() );
        if( c2.maybe('l') || c2.maybe('|') )
          {
          only_guess( 0, b1.left() ); add_guess( 'f', b2.left() );
          add_guess( 'i', b2.right() ); return;
          }
        }
      if( 2 * b1.height() > 3 * b2.height() &&
          Ocrad::similar( b2.width(), b2.height(), 50 ) )
        { add_guess( '%', 0 ); return; }
      }
    }
  if( a1 < a2 )
    {
      {
      int code = f1.test_solid( charbox_vcenter );	//FIXME all this
      if( code == '-' && 2 * b1.height() > b1.width() ) code = '.';
      else if( code == '\'' || code == '|' ) code = '.';
      if( !code && !b1.blocks() && 
          2 * b1.height() < b2.height() && b1.width() <= b2.width() )
        {
        if( 10 * a1 >= 7 * b1.height() * b1.width() ) code = '.';
        else code = '\'';
        }
      if( !b2.blocks() && ( code == '.' || code == '\'' ) )
        {
        // Looks for merged 'ri' or 'r'
        if( f2.bp.minima( b2.height() / 4 ) == 2 &&
            b2.top() > b1.bottom() && b2.hcenter() < b1.left() )
          {
          Character c2( b2 ); c2.recognize1( charset, vcenter() );
          if( c2.maybe('n') )
            {
            int col, limit = b2.seek_right( b2.vcenter(), b2.hcenter() );
            for( col = b2.hcenter(); col <= limit; ++col )
              if( b2.seek_bottom( b2.vcenter(), col ) < b2.bottom() ) break;
            if( b2.left() < col && col < b2.right() )
              {
              only_guess( 0, b2.left() );
              if( charset.enabled( Charset::iso_8859_9 ) && f2.rp.istip() )
                { add_guess( 'T', col - 1 ); add_guess( UCS::CIDOT, b2.right() ); }
              else
                {
                add_guess( 'r', col - 1 );
                if( code == '.' ) add_guess( 'i', b2.right() );
                else add_guess( UCS::SIACUTE, b2.right() );
                }
              return;
              }
            }
          }

        if( code == '.' && f2.bp.minima( b2.height() / 4 ) == 1 &&
            b1.bottom() <= b2.top() )
          {
          int hdiff;
          if( b2.bottom_hook( &hdiff ) && std::abs( hdiff ) >= b2.height() / 2 )
            {
            if( hdiff > 0 && f2.rp.increasing( f2.rp.pos( 75 ) ) )
              { add_guess( 'j', 0 ); return; }
            if( hdiff < 0 )
              {
              if( charset.enabled( Charset::iso_8859_15 ) ||
                  charset.enabled( Charset::iso_8859_9 ) )
                if( f2.wp.max() > 2 * f1.wp.max() && f2.lp.minima() == 1 )
                  { add_guess( UCS::IQUEST, 0 ); return; }
              add_guess( 'i', 0 ); return;
              }
            }
          if( f2.tp.minima() == 1 )
            {
            if( Ocrad::similar( f1.wp.max(), f2.wp.max(), 20 ) )
              {
              if( charset.enabled( Charset::iso_8859_15 ) ||
                  charset.enabled( Charset::iso_8859_9 ) )
                if( f2.wp.max() >= f1.wp.max() &&
                    3 * f2.wp[f2.wp.pos(10)] < 2 * f1.wp.max() )
                  { add_guess( UCS::IEXCLAM, 0 ); return; }
              add_guess( 'i', 0 ); return;
              }
            if( 3 * f2.wp.max() > 4 * f1.wp.max() &&
                b2.seek_bottom( b2.vcenter(), b2.hpos( 10 ) ) < b2.bottom() &&
                f2.rp.increasing( f2.rp.pos( 75 ) ) )
              { add_guess( 'j', 0 ); return; }
            if( charset.enabled( Charset::iso_8859_9 ) && f2.rp.istip() )
              { add_guess( UCS::CIDOT, 0 ); return; }
            add_guess( 'i', 0 ); return;
            }
          }
        }
      }

    // upper block has no holes
    if( ( b1.bottom() < b2.vcenter() || 2 * a1 < a2 ) && !b1.blocks() )
      {
      Character c( b2 ); c.recognize1( charset, c.vcenter() );
      if( c.guesses() )
        {
        int code = c.guess( 0 ).code;
        if( b1.bottom() < b2.vcenter() )
          {
          int atype = '\'';
          if( UCS::isvowel( code ) &&
              b1.width() > 2 * b1.height() && f1.hp.iscpit() )
            atype = ':';
          else if( f1.bp.minima() == 2 ) atype = '^';
          else
            {
            int slope; if( f1.rp.straight( &slope ) && slope < 0 ) atype = '`';
            }
          code = UCS::compose( code, atype );
          if( charset.only( Charset::ascii ) )
            {
            if( UCS::base_letter( code ) == 'i' ) code = 'i';
            else code = c.guess( 0 ).code;
            }
          }
        if( code ) add_guess( code, 0 );
        }
      }
    return;
    }

  if( b1.bottom() <= b2.top() )
  {
  int code = f2.test_solid( charbox_vcenter );
  if( !b1.blocks() && ( code == '.' || 
        ( code && Ocrad::similar( b2.height(), b2.width(), 50 ) ) ) )
    {
    if( Ocrad::similar( b1.width(), b2.width(), 50 ) )
      { add_guess( '!', 0 ); return; }
    add_guess( '?', 0 ); return;
    }
  if( code == '-' || code == '_' )
    if( charset.enabled( Charset::iso_8859_15 ) ||
        charset.enabled( Charset::iso_8859_9 ) )
      {
      if( b1.blocks() == 1 )
        {
        const Block & h = b1.block( 0 );
        if( Ocrad::similar( h.left() - b1.left(), b1.right() - h.right(), 40 ) )
          { add_guess( UCS::MASCORD, 0 ); return; }
        add_guess( UCS::FEMIORD, 0 ); return;
        }
      }
  }
  }

==> ocrad-0.10.tar.bz2: ocrad-0.10/character_r12.cc <==
/*  GNU Ocrad - Optical Character Recognition program
    Copyright (C) 2003, 2004 Antonio Diaz Diaz.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <cstdio>
#include <vector>
#include "common.h"
#include "rectangle.h"
#include "ucs.h"
#include "block.h"
#include "blockmap.h"
#include "character.h"
#include "profile.h"
#include "feats.h"


// Recognizes 2 block characters.
// ij!%:;=?|
void Character::recognize12( const Charset & charset, int charbox_vcenter ) throw()
  {
  const Block & b1 = block( 0 );		// upper block
  const Block & b2 = block( 1 );		// lower block
  int a1 = b1.area();
  int a2 = b2.area();
  Features f1( b1 );
  Features f2( b2 );

  if( Ocrad::similar( a1, a2, 10 ) )
    {
    if( !b1.blocks() && !b2.blocks() &&
        2 * a1 > b1.size() && 2 * a2 > b2.size() )
      {
      if( width() > height() || Ocrad::similar( width(), height(), 40 ) )
        { add_guess( '=', 0 ); return; }
      if( Ocrad::similar( b1.width(), b1.height(), 20, 2 ) &&
          Ocrad::similar( b2.width(), b2.height(), 20, 2 ) )
        add_guess( ':', 0 );
      return;
      }
    return;
    }
  if( Ocrad::similar( a1, a2, 60 ) )
    {
    if( f1.test_solid( charbox_vcenter ) == '.' )
      {
      if( f2.test_solid( charbox_vcenter ) == '.' )
        { add_guess( ':', 0 ); return; }
      if( b2.height() > b1.height() && b2.top() > charbox_vcenter )
        { add_guess( ';', 0 ); return; }
      }
    if( charset.enabled( Charset::iso_8859_15 ) ||
        charset.enabled( Charset::iso_8859_9 ) )
      {
      int code = f2.test_solid( charbox_vcenter );
      if( code == '-' || code == '_' )
        { add_guess( UCS::PLUSMIN, 0 ); return; }
      }
    if( b1.includes_hcenter( b2 ) && b2.includes_hcenter( b1 ) )
      { if( b1.blocks() && b2.blocks() ) { add_guess( 'g', 0 ); return; } }
    if( b1.hcenter() < b2.hcenter() )
      {
      // Looks for merged 'fi'
      if( b2.height() > b2.width() && b1.hcenter() < b2.left() &&
          b1.includes_hcenter( b2 ) && 4 * b1.height() > 5 * b2.height() &&
          Ocrad::similar( b1.bottom()-b1.top(), b2.bottom()-b1.top(), 10 ) )
        {
        Character c2( b2 ); c2.recognize1( charset, vcenter() );
        if( c2.maybe('l') || c2.maybe('|') )
          {
          only_guess( 0, b1.left() ); add_guess( 'f', b2.left() );
          add_guess( 'i', b2.right() ); return;
          }
        }
      if( 2 * b1.height() > 3 * b2.height() &&
          Ocrad::similar( b2.width(), b2.height(), 50 ) )
        { add_guess( '%', 0 ); return; }
      }
    }
  if( a1 < a2 )
    {
      {
      int code = f1.test_solid( charbox_vcenter );	//FIXME all this
      if( code == '-' && 2 * b1.height() > b1.width() ) code = '.';
      else if( code == '\'' || code == '|' ) code = '.';
      if( !code && !b1.blocks() && 
          2 * b1.height() < b2.height() && b1.width() <= b2.width() )
        {
        if( 10 * a1 >= 7 * b1.height() * b1.width() ) code = '.';
        else code = '\'';
        }
      if( !b2.blocks() && ( code == '.' || code == '\'' ) )
        {
        // Looks for merged 'ri' or 'r'
        if( f2.bp.minima( b2.height() / 4 ) == 2 &&
            b2.top() > b1.bottom() && b2.hcenter() < b1.left() )
          {
          Character c2( b2 ); c2.recognize1( charset, vcenter() );
          if( c2.maybe('n') )
            {
            int col, limit = b2.seek_right( b2.vcenter(), b2.hcenter() );
            for( col = b2.hcenter(); col <= limit; ++col )
              if( b2.seek_bottom( b2.vcenter(), col ) < b2.bottom() ) break;
            if( b2.left() < col && col < b2.right() )
              {
              only_guess( 0, b2.left() );
              if( charset.enabled( Charset::iso_8859_9 ) && f2.rp.istip() )
                { add_guess( 'T', col - 1 ); add_guess( UCS::CIDOT, b2.right() ); }
              else
                {
                add_guess( 'r', col - 1 );
                if( code == '.' ) add_guess( 'i', b2.right() );
                else add_guess( UCS::SIACUTE, b2.right() );
                }
              return;
              }
            }
          }

        if( code == '.' && f2.bp.minima( b2.height() / 4 ) == 1 &&
            b1.bottom() <= b2.top() )
          {
          int hdiff;
          if( b2.bottom_hook( &hdiff ) && std::abs( hdiff ) >= b2.height() / 2 )
            {
            if( hdiff > 0 && f2.rp.increasing( f2.rp.pos( 75 ) ) )
              { add_guess( 'j', 0 ); return; }
            if( hdiff < 0 )
              {
              if( charset.enabled( Charset::iso_8859_15 ) ||
                  charset.enabled( Charset::iso_8859_9 ) )
                if( f2.wp.max() > 2 * f1.wp.max() && f2.lp.minima() == 1 )
                  { add_guess( UCS::IQUEST, 0 ); return; }
              add_guess( 'i', 0 ); return;
              }
            }
          if( f2.tp.minima() == 1 )
            {
            if( Ocrad::similar( f1.wp.max(), f2.wp.max(), 20 ) )
              {
              if( charset.enabled( Charset::iso_8859_15 ) ||
                  charset.enabled( Charset::iso_8859_9 ) )
                if( !f2.lp.isctip() && f2.wp.max() >= f1.wp.max() &&
                    3 * f2.wp[f2.wp.pos(10)] < 2 * f1.wp.max() )
                  { add_guess( UCS::IEXCLAM, 0 ); return; }
              add_guess( 'i', 0 ); return;
              }
            if( 3 * f2.wp.max() > 4 * f1.wp.max() &&
                b2.seek_bottom( b2.vcenter(), b2.hpos( 10 ) ) < b2.bottom() &&
                f2.rp.increasing( f2.rp.pos( 75 ) ) &&
                ( b1.left() >= b2.hcenter() ||
                  b2.seek_top( b2.vcenter(), b2.hpos( 10 ) ) <= b2.top() ) )
              { add_guess( 'j', 0 ); return; }
            if( charset.enabled( Charset::iso_8859_9 ) && f2.rp.istip() )
              { add_guess( UCS::CIDOT, 0 ); return; }
            add_guess( 'i', 0 ); return;
            }
          }
        }
      }

    // upper block has no holes
    if( ( b1.bottom() < b2.vcenter() || 2 * a1 < a2 ) && !b1.blocks() )
      {
      Character c( b2 ); c.recognize1( charset, c.vcenter() );
      if( c.guesses() )
        {
        int code = c.guess( 0 ).code;
        if( b1.bottom() < b2.vcenter() )
          {
          int atype = '\'';
          if( UCS::isvowel( code ) &&
              b1.width() > 2 * b1.height() && f1.hp.iscpit() )
            atype = ':';
          else if( f1.bp.minima() == 2 ) atype = '^';
          else
            {
            int slope; if( f1.rp.straight( &slope ) && slope < 0 ) atype = '`';
            }
          code = UCS::compose( code, atype );
          if( charset.only( Charset::ascii ) )
            {
            if( UCS::base_letter( code ) == 'i' ) code = 'i';
            else code = c.guess( 0 ).code;
            }
          }
        if( code ) add_guess( code, 0 );
        }
      }
    return;
    }

  if( b1.bottom() <= b2.top() )
  {
  int code = f2.test_solid( charbox_vcenter );
  if( !b1.blocks() && ( code == '.' || 
        ( code && Ocrad::similar( b2.height(), b2.width(), 50 ) ) ) )
    {
    if( Ocrad::similar( b1.width(), b2.width(), 50 ) )
      { add_guess( '!', 0 ); return; }
    add_guess( '?', 0 ); return;
    }
  if( code == '-' || code == '_' )
    if( charset.enabled( Charset::iso_8859_15 ) ||
        charset.enabled( Charset::iso_8859_9 ) )
      {
      if( b1.blocks() == 1 )
        {
        const Block & h = b1.block( 0 );
        if( Ocrad::similar( h.left() - b1.left(), b1.right() - h.right(), 40 ) )
          { add_guess( UCS::MASCORD, 0 ); return; }
        add_guess( UCS::FEMIORD, 0 ); return;
        }
      }
  }
  }
